@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 
 # ----------------------------------------------------------------------------
-# Copyright (c) 2011-2013, The BIOM Format Development Team.
+# Copyright (c) 2011-2017, The BIOM Format Development Team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -9,16 +9,19 @@
 # ----------------------------------------------------------------------------
 
 from __future__ import division
-from string import maketrans
+
 import numpy as np
-from biom import __version__
+from future.utils import string_types
+
 from biom.exception import BiomParseException, UnknownAxisError
 from biom.table import Table
-from biom.util import biom_open
+from biom.util import biom_open, __version__
 import json
+from collections import defaultdict, OrderedDict
+
 
 __author__ = "Justin Kuczynski"
-__copyright__ = "Copyright 2011-2013, The BIOM Format Development Team"
+__copyright__ = "Copyright 2011-2017, The BIOM Format Development Team"
 __credits__ = ["Justin Kuczynski", "Daniel McDonald", "Greg Caporaso",
                "Jose Carlos Clemente Litran", "Adam Robbins-Pianka",
                "Jose Antonio Navas Molina"]
@@ -27,8 +30,8 @@ __url__ = "http://biom-format.org"
 __maintainer__ = "Daniel McDonald"
 __email__ = "daniel.mcdonald@colorado.edu"
 
-MATRIX_ELEMENT_TYPE = {'int': int, 'float': float, 'unicode': unicode,
-                       u'int': int, u'float': float, u'unicode': unicode}
+MATRIX_ELEMENT_TYPE = {'int': int, 'float': float, 'unicode': str,
+                       u'int': int, u'float': float, u'unicode': str}
 
 QUOTE = '"'
 JSON_OPEN = set(["[", "{"])
@@ -125,7 +128,7 @@ def direct_slice_data(biom_str, to_keep, axis):
 
     # determine shape
     raw_shape = shape_kv_pair.split(':')[-1].replace("[", "").replace("]", "")
-    n_rows, n_cols = map(int, raw_shape.split(","))
+    n_rows, n_cols = list(map(int, raw_shape.split(",")))
 
     # slice to just data
     data_start = data_fields.find('[') + 1
@@ -157,18 +160,20 @@ def direct_slice_data(biom_str, to_keep, axis):
 
     return '"data": %s, "shape": %s' % (new_data, new_shape)
 
-STRIP_F = lambda x: x.strip("[] \n\t")
+
+def strip_f(x):
+    return x.strip("[] \n\t")
 
 
 def _remap_axis_sparse_obs(rcv, lookup):
     """Remap a sparse observation axis"""
-    row, col, value = map(STRIP_F, rcv.split(','))
+    row, col, value = list(map(strip_f, rcv.split(',')))
     return "%s,%s,%s" % (lookup[row], col, value)
 
 
 def _remap_axis_sparse_samp(rcv, lookup):
     """Remap a sparse sample axis"""
-    row, col, value = map(STRIP_F, rcv.split(','))
+    row, col, value = list(map(strip_f, rcv.split(',')))
     return "%s,%s,%s" % (row, lookup[col], value)
 
 
@@ -182,7 +187,7 @@ def _direct_slice_data_sparse_obs(data, to_keep):
     new_data = []
     remap_lookup = dict([(str(v), i) for i, v in enumerate(sorted(to_keep))])
     for rcv in data.split('],'):
-        r, c, v = STRIP_F(rcv).split(',')
+        r, c, v = strip_f(rcv).split(',')
         if r in remap_lookup:
             new_data.append(_remap_axis_sparse_obs(rcv, remap_lookup))
     return '[[%s]]' % '],['.join(new_data)
@@ -244,6 +249,98 @@ def get_axis_indices(biom_str, to_keep, axis):
     return idxs, json.dumps(subset)[1:-1]  # trim off { and }
 
 
+def parse_uc(fh):
+    """ Create a Table object from a uclust/usearch/vsearch uc file.
+
+        Parameters
+        ----------
+        fh : file handle
+            The ``.uc`` file to be parsed.
+
+        Returns
+        -------
+        biom.Table : The resulting BIOM table.
+
+        Raises
+        ------
+        ValueError
+            If a sequence identifier is encountered that doesn't have at least
+            one underscore in it (see Notes).
+
+        Notes
+        -----
+        This function assumes sequence identifiers in this file are in QIIME's
+        "post-split-libraries" format, where the identifiers are of the form
+        ``<sample-id>_<sequence-id>``. Everything before the first underscore
+        will be used as the sample identifier in the resulting ``Table``.
+        The information after the first underscore is not used directly, though
+        the full identifiers of seeds will be used as the observation
+        identifier in the resulting ``Table``.
+
+    """
+    data = defaultdict(int)
+    sample_idxs = {}
+    sample_ids = []
+    observation_idxs = {}
+    observation_ids = []
+    # The types of hit lines we need here are hit (H), seed (S) and
+    # library seed (L). Store these in a set for quick reference.
+    line_types = set('HSL')
+    for line in fh:
+        # determine if the current line is one that we need
+        line = line.strip()
+        if not line:
+            continue
+        fields = line.split('\t')
+
+        line_type = fields[0]
+        if line_type not in line_types:
+            continue
+
+        # grab the fields we care about
+        observation_id = fields[9].split()[0]
+        query_id = fields[8].split()[0]
+
+        if observation_id == '*':
+            # S and L lines don't have a separate observation id
+            observation_id = query_id
+
+        # get the index of the current observation id, or create it if it's
+        # the first time we're seeing this id
+        if observation_id in observation_idxs:
+            observation_idx = observation_idxs[observation_id]
+        else:
+            observation_idx = len(observation_ids)
+            observation_ids.append(observation_id)
+            observation_idxs[observation_id] = observation_idx
+
+        if line_type == 'H' or line_type == 'S':
+            # get the sample id
+            try:
+                underscore_index = query_id.rindex('_')
+            except ValueError:
+                raise ValueError(
+                 "A query sequence was encountered that does not have an "
+                 "underscore. An underscore is required in all query "
+                 "sequence identifiers to indicate the sample identifier.")
+            # get the sample id and its index, creating the index if it is the
+            # first time we're seeing this id
+            sample_id = query_id[:underscore_index]
+            if sample_id in sample_idxs:
+                sample_idx = sample_idxs[sample_id]
+            else:
+                sample_idx = len(sample_ids)
+                sample_idxs[sample_id] = sample_idx
+                sample_ids.append(sample_id)
+            # increment the count of the current observation in the current
+            # sample by one.
+            data[(observation_idx, sample_idx)] += 1
+        else:
+            # nothing else needs to be done for 'L' records
+            pass
+    return Table(data, observation_ids=observation_ids, sample_ids=sample_ids)
+
+
 def parse_biom_table(fp, ids=None, axis='sample', input_is_dense=False):
     r"""Parses the biom table stored in the filepath `fp`
 
@@ -295,31 +392,45 @@ def parse_biom_table(fp, ids=None, axis='sample', input_is_dense=False):
 
     try:
         return Table.from_hdf5(fp, ids=ids, axis=axis)
-    except:
+    except ValueError:
+        pass
+    except RuntimeError:
         pass
-
     if hasattr(fp, 'read'):
         old_pos = fp.tell()
-        try:
-            t = Table.from_json(json.load(fp), input_is_dense=input_is_dense)
-        except ValueError:
+        # Read in characters until first non-whitespace
+        # If it is a {, then this is (most likely) JSON
+        c = fp.read(1)
+        while c.isspace():
+            c = fp.read(1)
+        if c == '{':
+            fp.seek(old_pos)
+            t = Table.from_json(json.load(fp, object_pairs_hook=OrderedDict),
+                                input_is_dense=input_is_dense)
+        else:
             fp.seek(old_pos)
             t = Table.from_tsv(fp, None, None, lambda x: x)
     elif isinstance(fp, list):
         try:
-            t = Table.from_json(json.loads(''.join(fp)),
+            t = Table.from_json(json.loads(''.join(fp),
+                                           object_pairs_hook=OrderedDict),
                                 input_is_dense=input_is_dense)
         except ValueError:
             t = Table.from_tsv(fp, None, None, lambda x: x)
     else:
-        t = Table.from_json(json.loads(fp), input_is_dense=input_is_dense)
+        t = Table.from_json(json.loads(fp, object_pairs_hook=OrderedDict),
+                            input_is_dense=input_is_dense)
+
+    def subset_ids(data, id_, md):
+        return id_ in ids
+
+    def gt_zero(vals, id_, md):
+        return np.any(vals)
 
     if ids is not None:
-        f = lambda data, id_, md: id_ in ids
-        t.filter(f, axis=axis)
+        t.filter(subset_ids, axis=axis)
         axis = 'observation' if axis == 'sample' else 'sample'
-        f = lambda vals, id_, md: np.any(vals)
-        t.filter(f, axis=axis)
+        t.filter(gt_zero, axis=axis)
 
     return t
 
@@ -371,18 +482,22 @@ class MetadataMap(dict):
 
         if strip_quotes:
             if suppress_stripping:
-                # remove quotes but not spaces
-                strip_f = lambda x: x.replace('"', '')
+                def strip_f(x):
+                    # remove quotes but not spaces
+                    return x.replace('"', '')
             else:
-                # remove quotes and spaces
-                strip_f = lambda x: x.replace('"', '').strip()
+                def strip_f(x):
+                    # remove quotes and spaces
+                    return x.replace('"', '').strip()
         else:
             if suppress_stripping:
-                # don't remove quotes or spaces
-                strip_f = lambda x: x
+                def strip_f(x):
+                    # don't remove quotes or spaces
+                    return x
             else:
-                # remove spaces but not quotes
-                strip_f = lambda x: x.strip()
+                def strip_f(x):
+                    # remove spaces but not quotes
+                    return x.strip()
 
         # if the user didn't provide process functions, initialize as
         # an empty dict
@@ -409,7 +524,7 @@ class MetadataMap(dict):
                     comments.append(line)
             else:
                 # Will add empty string to empty fields
-                tmp_line = map(strip_f, line.split('\t'))
+                tmp_line = list(map(strip_f, line.split('\t')))
                 if len(tmp_line) < len(header):
                     tmp_line.extend([''] * (len(header) - len(tmp_line)))
                 mapping_data.append(tmp_line)
@@ -473,11 +588,10 @@ def biom_meta_to_string(metadata, replace_str=':'):
     # Note that since ';' and '|' are used as seperators we must replace them
     # if they exist
 
-    # metadata is just a string (not a list)
-    if isinstance(metadata, str) or isinstance(metadata, unicode):
+    if isinstance(metadata, string_types):
         return metadata.replace(';', replace_str)
     elif isinstance(metadata, list):
-        transtab = maketrans(';|', ''.join([replace_str, replace_str]))
+        transtab = bytes.maketrans(';|', ''.join([replace_str, replace_str]))
         # metadata is list of lists
         if isinstance(metadata[0], list):
             new_metadata = []
